/*************************************************************************************
 *  Copyright (C) 2012 by Alejandro Fiestas Olivares <afiestas@kde.org>              *
 *  Copyright (C) 2014 by Daniel Vrátil <dvratil@redhat.com>                         *
 *                                                                                   *
 *  This library is free software; you can redistribute it and/or                    *
 *  modify it under the terms of the GNU Lesser General Public                       *
 *  License as published by the Free Software Foundation; either                     *
 *  version 2.1 of the License, or (at your option) any later version.               *
 *                                                                                   *
 *  This library is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                *
 *  Lesser General Public License for more details.                                  *
 *                                                                                   *
 *  You should have received a copy of the GNU Lesser General Public                 *
 *  License along with this library; if not, write to the Free Software              *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA       *
 *************************************************************************************/

#ifndef OUTPUT_CONFIG_H
#define OUTPUT_CONFIG_H

#include "mode.h"
#include "types.h"
#include "kscreen_export.h"

#include <QObject>
#include <QSize>
#include <QPoint>
#include <QMetaType>
#include <QStringList>
#include <QDebug>

namespace KScreen {

class Edid;

class KSCREEN_EXPORT Output : public QObject
{
    Q_OBJECT

    public:
        Q_ENUMS(Rotation)
        Q_ENUMS(Type)
        Q_PROPERTY(int id READ id CONSTANT)
        Q_PROPERTY(QString name READ name WRITE setName NOTIFY outputChanged)
        Q_PROPERTY(Type type READ type WRITE setType NOTIFY outputChanged)
        Q_PROPERTY(QString icon READ icon WRITE setIcon NOTIFY outputChanged)
        Q_PROPERTY(ModeList modes READ modes NOTIFY modesChanged)
        Q_PROPERTY(QPoint pos READ pos WRITE setPos NOTIFY posChanged)
        Q_PROPERTY(QSize size READ size WRITE setSize NOTIFY sizeChanged)
        Q_PROPERTY(Rotation rotation READ rotation WRITE setRotation NOTIFY rotationChanged)
        Q_PROPERTY(QString currentModeId READ currentModeId WRITE setCurrentModeId NOTIFY currentModeIdChanged)
        Q_PROPERTY(QString preferredModeId READ preferredModeId CONSTANT)
        Q_PROPERTY(bool connected READ isConnected WRITE setConnected NOTIFY isConnectedChanged)
        Q_PROPERTY(bool enabled READ isEnabled WRITE setEnabled NOTIFY isEnabledChanged)
        Q_PROPERTY(bool primary READ isPrimary WRITE setPrimary NOTIFY isPrimaryChanged)
        Q_PROPERTY(QList<int> clones READ clones WRITE setClones NOTIFY clonesChanged)
        Q_PROPERTY(int replicationSource READ replicationSource
                   WRITE setReplicationSource NOTIFY replicationSourceChanged)
        Q_PROPERTY(KScreen::Edid* edid READ edid CONSTANT)
        Q_PROPERTY(QSize sizeMm READ sizeMm CONSTANT)
        Q_PROPERTY(qreal scale READ scale WRITE setScale NOTIFY scaleChanged)
        Q_PROPERTY(QPoint middlePos READ middlePos WRITE setMiddlePos NOTIFY middlePosChanged)
        Q_PROPERTY(qreal xdpi READ xdpi WRITE setXdpi NOTIFY xdpiChanged)
        Q_PROPERTY(bool followPreferredMode READ followPreferredMode WRITE setFollowPreferredMode NOTIFY followPreferredModeChanged)
        Q_PROPERTY(QSizeF logicalSize READ logicalSize WRITE setLogicalSize
                   NOTIFY logicalSizeChanged)


        enum Type {
            Unknown,
            VGA,
            DVI,
            DVII,
            DVIA,
            DVID,
            HDMI,
            Panel,
            TV,
            TVComposite,
            TVSVideo,
            TVComponent,
            TVSCART,
            TVC4,
            DisplayPort
        };

        enum Rotation {
            None = 1,
            Left = 2,
            Inverted = 4,
            Right = 8
        };

        explicit Output();
        ~Output() override;

        OutputPtr clone() const;

        int id() const;
        void setId(int id);

        QString name() const;
        void setName(const QString& name);

        /**
         * Returns an identifying hash for this output.
         * 返回此输出的标识哈希。
         * The hash is calculated either via the edid hash or if no
         * edid is available by the output name.
         * 通过edid散列计算散列，或者如果没有可用的edid，则通过输出名称计算散列。
         * @return identifying hash of this output
         * @since 5.15
         * @deprecated
         * @see hashMd5
         */
        QString hash() const;

        /**
         * Returns an identifying hex encoded MD5-hash for this output.
         * 返回此输出的标识十六进制编码MD5哈希。
         * The hash is calculated either via the edid hash or if no
         * edid is available by the output name, which is hashed as well.
         * 散列可以通过edid散列计算，如果没有可用的edid，则通过输出名称计算，输出名称也会被散列。
         * @return identifying hash of this output 标识此输出的哈希
         * @since 5.17
         */
        QString hashMd5() const;

        Type type() const;
        void setType(Type type);

        QString icon() const;
        void setIcon(const QString& icon);

        Q_INVOKABLE ModePtr mode(const QString &id) const;
        ModeList modes() const;
        void setModes(const ModeList &modes);

        QString currentModeId() const;
        void setCurrentModeId(const QString& mode);
        Q_INVOKABLE ModePtr currentMode() const;

        void setPreferredModes(const QStringList &modes);
        QStringList preferredModes() const;
        /**
         * Returns the preferred mode with higher resolution and refresh
         */
        Q_INVOKABLE QString preferredModeId() const;
        /**
         * Returns KScreen::Mode associated with preferredModeId()
         */
        Q_INVOKABLE ModePtr preferredMode() const;

        QPoint pos() const;
        void setPos(const QPoint& pos);

        /***
         * Returns actual size being rendered in the output
         * 返回输出中呈现的实际大小
         * The returned valued is after transformations have been applied to
         * the resolution of the current mode.
         * 将转换应用于当前模式的分辨率后，返回值。
         * For example if currentMode is 1280x800 but it is a vertical screen
         * the returned size will be 800x1280.
         * 例如，如果currentMode为1280x800，但它是垂直屏幕，则返回的大小将为800x1280。
         * If that same resolution (1280x800) is transformed and scale x2, the
         * value returned will be 2560x1600.
         * 如果转换相同的分辨率（1280x800）并缩放x2，则返回的值将为2560x1600。
         * This property reflects the currently active output configuration and
         * is not affected by current mode or orientation change made by user
         * until the config is applied.
         * 此属性反映当前活动的输出配置，在应用配置之前，不受当前模式或用户更改方向的影响。
         * @since 5.4
         */
        QSize size() const;
        void setSize(const QSize& size);

        /**
         * Returns either current mode size or if not available preferred one or if also not
         * available the first one in the ModeList.
         * 返回当前模式大小，如果不可用，则返回首选模式，如果不可用，则返回ModeList中的第一个模式。
         * @return mode size
         */
        QSize enforcedModeSize() const;

        Rotation rotation() const;
        void setRotation(Rotation rotation);
        /**
         * A comfortable function that returns true when output is not rotated
         * or is rotated upside down.
         * 一个舒适的函数，当输出未旋转或倒置时返回true。
         */
        Q_INVOKABLE inline bool isHorizontal() const {
            return ((rotation() == Output::None) || (rotation() == Output::Inverted));
        }

        bool isConnected() const;
        void setConnected(bool connected);

        bool isEnabled() const;
        void setEnabled(bool enabled);

        bool isPrimary() const;
        void setPrimary(bool primary);

        /**
         * @brief Immutable clones because of hardware restrictions
         * 由于硬件限制而不可变克隆
         * Clones are set symmetcally on all outputs. The list contains ids
         * for all other outputs being clones of this output.
         * 克隆在所有输出上对称设置。该列表包含作为该输出克隆的所有其他输出的ID。
         * @return List of output ids being clones of each other.
         * 彼此克隆的输出ID列表。
         */
        QList<int> clones() const;
        /**
         * @brief Set the clones list.
         * 设置克隆列表。
         * When this output is part of a configuration this call is followed by
         * similar calls on other outputs making the lists in all outputs
         * symmetric.
         * 当此输出是配置的一部分时，此调用之后会对其他输出进行类似调用，从而使所有输出中的列表对称。
         * @param outputlist
         */
        void setClones(const QList<int> &outputlist);

        /**
         * @brief Provides the source for an ongoing replication
         * 提供正在进行的复制的源
         * If the returned output id is non-null this output is a replica of the
         * returned output. If null is returned the output is no replica of any
         * other output.
         * 如果返回的输出id为非空，则此输出是返回的输出的副本。如果返回null，则输出不是任何其他输出的副本。
         * @return Replication source output id of this output
         * 此输出的复制源输出id
         */
        int replicationSource() const;
        /**
         * @brief Set the replication source.
         * 设置复制源。
         * @param source
         */
        void setReplicationSource(int source);

        void setEdid(const QByteArray &rawData);

        /**
         * edid returns the output's EDID information if available.
         * edid返回输出的edid信息（如果可用）。
         * The output maintains ownership of the returned Edid, so the caller should not delete it.
         * Note that the edid is only valid as long as the output is alive.
         * 输出维护返回的Edid的所有权，因此调用者不应该删除它。请注意，只有当输出处于活动状态时，edid才有效。
         */
        Edid* edid() const;

        /**
         * Returns the physical size of the screen in milimeters.
         * 返回屏幕的物理大小（以毫米为单位）。
         * @note Some broken GPUs or monitors return the size in centimeters instead
         * of millimeters. KScreen at the moment is not sanitizing the values.
         * 一些损坏的GPU或显示器返回的大小是厘米，而不是毫米。KScreen目前没有对这些值进行消毒。
         */
        QSize sizeMm() const;
        void setSizeMm(const QSize &size);

        /**
         * Returns if the output needs to be taken account for in the overall compositor/screen
         * space and if it should be depicted on its own in a graphical view for repositioning.
         * 返回是否需要在整个合成器/屏幕空间中考虑输出，以及是否应该在图形视图中单独描述输出以重新定位。
         * @return true if the output is positionable in compositor/screen space.
         * 如果输出可在合成器/屏幕空间中定位，则为true。
         * @since 5.17
         */
        bool isPositionable() const;

        /**
         * Returns a rectangle containing the currently set output position and
         * size.
         * 返回包含当前设置的输出位置和大小的矩形。
         * The geometry also reflects current orientation (i.e. if current mode
         * is 1920x1080 and orientation is @p KScreen::Output::Left, then the
         * size of the returned rectangle will be 1080x1920.
         * 几何图形还反映当前方向（即，如果当前模式为1920x1080，方向为@p KScreen:：Output:：Left，则返回矩形的大小将为1080x1920）。
         * This property contains the current settings stored in the particular
         * Output object, so it is updated even when user changes current mode
         * or orientation without applying the whole config/
         * 此属性包含存储在特定输出对象中的当前设置，因此即使用户在不应用整个配置的情况下更改当前模式或方向，它也会更新/
         */
        QRect geometry() const;

        /**
         * returns the scaling factor to use for this output
         *
         * @since 5.9
         */
        qreal scale() const;

        /**
         * Set the scaling factor for this output.
         * 设置此输出的比例因子。
         * @arg factor Scale factor to use for this output, the backend may or may not
         * be able to deal with non-integer values, in that case, the factor gets rounded.
         * 要用于此输出，后端可能无法处理非整数值，在这种情况下，该因子将被舍入。
         * @since 5.9
         */
        void setScale(qreal factor);
        /**
         * returns the xdpi to use for this output
         *
         * @since 5.9
         */
        qreal xdpi() const;
        /**
         * returns the middlePos to use for this output
         *
         * @since 5.9
         */
        QPoint middlePos() const;

        /**
         * Set the scaling factor for this output.
         * 设置此输出的比例因子。
         * @arg factor Scale factor to use for this output, the backend may or may not
         * be able to deal with non-integer values, in that case, the factor gets rounded.
         * 要用于此输出，后端可能无法处理非整数值，在这种情况下，该因子将被舍入。
         * @since 5.9
         */
        void setXdpi(qreal factor);
        /**
         * Set the middlePos for this output.
         * 设置此输出的居中坐标。
         * @arg factor Scale factor to use for this output, the backend may or may not
         * be able to deal with non-integer values, in that case, the factor gets rounded.
         * 要用于此输出，后端可能无法处理非整数值，在这种情况下，该因子将被舍入。
         * @since 5.9
         */
        void setMiddlePos(QPoint middlePos);
        /**
         * The logical size is the output's representation internal to the display server and its
         * overall screen geometry.
         * 逻辑大小是显示服务器内部的输出表示及其总体屏幕几何图形。返回此输出的逻辑大小
         * returns the logical size of this output
         *
         * @since 5.18
         */
        QSizeF logicalSize() const;

        /**
         * The logical size is the output's representation internal to the display server and its
         * overall screen geometry.
         * 逻辑大小是显示服务器内部的输出表示及其总体屏幕几何图形。返回此输出的显式设置的逻辑大小，如果未设置，则该大小无效
         * returns the explicitly set logical size of this output, is an invalid size if not set
         *
         * @since 5.18
         */
        QSizeF explicitLogicalSize() const;

        /**
         * Specifies explicitly the logical size of this output and by that overrides any other
         * logical size calculation through mode and scale. To enable this instead again call this
         * function with an invalid size as argument.
         * 明确指定此输出的逻辑大小，并通过模式和比例覆盖任何其他逻辑大小计算。要启用此功能，请使用无效的大小作为参数再次调用此函数。
         * @param size of this output in logical space
         *
         * @since 5.18
         */
        void setLogicalSize(const QSizeF &size);

        /**
         * @returns whether the mode should be changed to the new preferred mode
         * once it changes
         *
         * @since 5.15
         */
        bool followPreferredMode() const;

        /**
         * Set whether the preferred mode should be followed through @arg follow
         *
         * @since 5.15
         */
        void setFollowPreferredMode(bool follow);

        void apply(const OutputPtr &other);
    Q_SIGNALS:
        void outputChanged();
        void posChanged();
        void sizeChanged();
        void currentModeIdChanged();
        void rotationChanged();
        void isConnectedChanged();
        void isEnabledChanged();
        void isPrimaryChanged();
        void clonesChanged();
        void replicationSourceChanged();
        void scaleChanged();
        void xdpiChanged();
        void middlePosChanged();
        void logicalSizeChanged();
        void followPreferredModeChanged(bool followPreferredMode);

        /** The mode list changed.
         *
         * This may happen when a mode is added or changed.
         *
         * @since 5.8.3
         */
        void modesChanged();

    private:
        Q_DISABLE_COPY(Output)

        class Private;
        Private * const d;

        Output(Private *dd);
};

} //KScreen namespace

KSCREEN_EXPORT QDebug operator<<(QDebug dbg, const KScreen::OutputPtr &output);

Q_DECLARE_METATYPE(KScreen::OutputList)
Q_DECLARE_METATYPE(KScreen::Output::Rotation)
Q_DECLARE_METATYPE(KScreen::Output::Type)



#endif //OUTPUT_H
